############################################################################
#                                                                          #
#   Copyright (C) 2011 Po-Hsien Lee <p.lee@mx.uni-saarland.de>             #
#   All rights reserved.                                                   #
#                                                                          #
#   This file is part of PROPORES.                                         #
#                                                                          #
#   PROPORES is free software: you can redistribute it and/or modify       #
#   it under the terms of the GNU General Public License as published by   #
#   the Free Software Foundation, either version 3 of the License, or      #
#   (at your option) any later version.                                    #
#                                                                          #
#   PROPORES is distributed in the hope that it will be useful,            #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of         #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
#   GNU General Public License for more details.                           #
#                                                                          #
#   You should have received a copy of the GNU General Public License      #
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.  #
#                                                                          #
############################################################################
#                                                                          #
#   Gate_Open.pl.pl                                                        #
#                                                                          #
#   Programmed by Po-Hsien Lee <p.lee@mx.uni-saarland.de>, 2011            #
#   Center for Bioinformatics, Saarland University, Germany                #
#                                                                          #
#   Gate_Open.pl is a program for opening the gate between two             #
#   neighboring pores by rotating side chains of the gating residues.      #
#                                                                          #
############################################################################

#!/usr/bin/perl

### include necessary perl module and initialization
use Parameter;
use PDBreader;
use Vector;
use POSIX;
use KD_tree_NN;
use Dihedral;
use Quat_rot;
use Getopt::Std;

$pdb_obj=PDBreader->new();
$vector_op=Vector->new();
$kd_tree_op=KD_tree_NN->new();
$quad_rot=Quat_rot->new();


### read in the rotamer libraries
$rotamer_lib_path=$ENV{'ROTALIB'};
if($rotamer_lib_path eq "")
{
	print "\nPlease source the EnvVar file to get enviromental variables\n\n";
	exit;
}
print "ROTALIB=$rotamer_lib_path\n";
$clash_tolerance_ga=1.0; # tolerance of vdw overlap for gamma C or O, the unit is angstrom
($Tuser,$Tsystem,$Tchild,$Tchildsystem)=times;


### get and check input parameters
($pdb_file, $gate_list,$clash_tolerance)=&get_check_options();


### read in pdb file and gate_residue information
@ATOM=();
$pdb_obj->pdbparser($pdb_file,\@ATOM);

%gate=(); $number_of_gating_residue=0;
open FH, "<$gate_list" or die "$! Cannot open the file";
while(<FH>)
{
	if(m/(.{1})\s+(\d+)\s+(\w{3})/)
	{
		$gate{$1}{$2}=$3;
		$number_of_gating_residue++;
	}
}
close FH;


@gate_ca=(); $gate_ca_count=0;
%Gate_Atom=(); %at_aryidx=(); $temp_count=0;
for($i=0; $i<=$#ATOM; $i++)
{

	$ATOM[$i][17]=$i; # specific for KD-tree construction

	$chain=$ATOM[$i][7];
	$resi=$ATOM[$i][8];
	$res_ty=$ATOM[$i][6];
	$atom_ty=$ATOM[$i][3];

	if( $gate{$chain}{$resi} eq $res_ty )
	{
		if($atom_ty eq "CA")
		{
			$gate_ca[0]+=$ATOM[$i][9];
			$gate_ca[1]+=$ATOM[$i][10];
			$gate_ca[2]+=$ATOM[$i][11];
			$gate_ca_count++;
		}

		if($res_ty ne "GLY" && $res_ty ne "ALA" && $res_ty ne "PRO")
		{
			$last_idx=$#{$Gate_Atom{$chain}{$resi}};
			@{$Gate_Atom{$chain}{$resi}[$last_idx+1]}=@{$ATOM[$i]};
			$at_aryidx{$chain}{$resi}{$atom_ty}=$last_idx+1;

			if($atom_ty ne 'N' && $atom_ty ne 'CA' && $atom_ty ne 'C' && $atom_ty ne 'O')
			{
				$ATOM[$i][9]=$ATOM[$i][10]=$ATOM[$i][11]=99999+$temp_count; # addition of temp_count is for preventing the unlimited iteration of KD tree construction !!!
				$temp_count++;
			}
		}
	}
}
$gate_ca[0]/=$gate_ca_count;
$gate_ca[1]/=$gate_ca_count;
$gate_ca[2]/=$gate_ca_count;
#print "$gate_ca[0]\t$gate_ca[1]\t$gate_ca[2]\n";
print "Step 1 : Read in pdb file and gate information : finished\n\n";


### filter out the rotamer with serious clash with all other atoms of the protein
### construct KD_tree of the protein truncated side chains of gating residues. 

($user,$system,$child,$childsystem)=times;
$tree={};
$tree=$kd_tree_op->construct_kd_tree(\@ATOM,0);
%rotamer_lib=(); %score_single=(); %side_chain_ref_chi=();
while(($key1)=each(%Gate_Atom))
{
	while(($key2)=each(%{$Gate_Atom{$key1}}))
	{
		$value=$Gate_Atom{$key1}{$key2}[0][6]; # residue type of this gate residue
		&side_chain_dihedral_angle($value,\@{$side_chain_ref_chi{$key1}{$key2}},\@{$Gate_Atom{$key1}{$key2}},\%{$at_aryidx{$key1}{$key2}});
	
		$rotamer_count=0; $check_count=0;
		open FH, "< $rotamer_lib_path/$value\_trimed.txt" or die "$!";
		while(<FH>)
		{
			$check_count++;
			$score_temp=0;
			@temp=@target_side_chain=();
			@temp=split(/\s+/,$_);
			&rotamer_generation($value, \@{$side_chain_ref_chi{$key1}{$key2}}, \%{$at_aryidx{$key1}{$key2}}, \@{$Gate_Atom{$key1}{$key2}}, \@temp, \@target_side_chain);
			($clash_flag, $score_temp)=&serious_clash(\@target_side_chain, $tree, \@ATOM, \@gate_ca);
			
			if($clash_flag==0)
			{
				$last_index=$#{$rotamer_lib{$key1}{$key2}};
				@{$rotamer_lib{$key1}{$key2}[$last_index+1]}=@temp;
				$score_single{$key1}{$key2}[$last_index+1]=$score_temp;
				#$pdb_obj->pdb_print("$key1\_$key2\_$check_count.pdb",\@target_side_chain);
			}
			elsif($clash_flag==1) # print for checking in molecular graphic software
			{
				$last_index=$#{$rotamer_lib{$key1}{$key2}};
				#$pdb_obj->pdb_print("$key1\_$key2\_$check_count\_C.pdb",\@target_side_chain);
			}
		}
		close FH;
	}
}
($user1,$system1,$child1,$childsystem1)=times;
$user1-=$user;
$system1-=$system;
$child1-=$child;
$childsystem1-=$childsystem;
print "Step 2 : Filter rotamer with serious clashes of backbone in the rotamer library : finished\n";
printf "running time = %.1f (s)\n\n", $user1;
#print "check_count=$check_count\n";



### Dead End Elimination for the best combination of side chain rotamers
($user,$system,$child,$childsystem)=times;

## sort list of gated_residue for improving a little bit of efficient of dead end elimination
@sorted_sequence=(); $count=0;
while(($key1)=each(%Gate_Atom))
{
	while(($key2)=each(%{$Gate_Atom{$key1}}))
	{
		$value=$Gate_Atom{$key1}{$key2}[0][6];

		$length=$#{$rotamer_lib{$key1}{$key2}};
		$sorted_sequence[$count][0]=$key1;
		$sorted_sequence[$count][1]=$key2;
		$sorted_sequence[$count][2]=$value;
		$sorted_sequence[$count][3]=$length;
		$count++;
	}
}
@sorted_sequence=sort{$a->[3] <=> $b->[3]}(@sorted_sequence);


### dead end elimination (Robert Goldstein's formula)
for( $i=0; $i<=$#sorted_sequence; $i++ )
{
	$key1=$sorted_sequence[$i][0];
	$key2=$sorted_sequence[$i][1];
	$value=$sorted_sequence[$i][2];
	$iteration=$#{$rotamer_lib{$key1}{$key2}};


	for($j=0; $j<=$iteration-1; $j++)
	{
		if($#{$rotamer_lib{$key1}{$key2}[$j]}==-1)
		{
			next;
		}
		@rotamer_1=();
		&rotamer_generation($value, \@{$side_chain_ref_chi{$key1}{$key2}}, \%{$at_aryidx{$key1}{$key2}}, \@{$Gate_Atom{$key1}{$key2}}, \@{$rotamer_lib{$key1}{$key2}[$j]}, \@rotamer_1);
		
		for($k=$j+1; $k<=$iteration; $k++)
		{
			if($#{$rotamer_lib{$key1}{$key2}[$k]}==-1)
			{
				next;
			}
			elsif($#{$rotamer_lib{$key1}{$key2}[$j]}==-1)
			{
				last;
			}

			@rotamer_2=();
			&rotamer_generation($value, \@{$side_chain_ref_chi{$key1}{$key2}}, \%{$at_aryidx{$key1}{$key2}}, \@{$Gate_Atom{$key1}{$key2}}, \@{$rotamer_lib{$key1}{$key2}[$k]}, \@rotamer_2);

			# $pdb_obj->pdb_print("$key1\_$key2\_$value\_$j.pdb",\@rotamer_1);
			($min_score_1M2, $min_score_2M1)=&dead_end_elimination($i, \@sorted_sequence, \@rotamer_1, \@rotamer_2, \%rotamer_lib, \%side_chain_ref_chi, \%at_aryidx, \%Gate_Atom);
			
			#print "$j $k min_score 1M2=$min_score_1M2\t2M1=$min_score_2M1\n";

			
			$backbone_1M2=$score_single{$key1}{$key2}[$j]-$score_single{$key1}{$key2}[$k];
			$backbone_2M1=-1*$backbone_1M2;
			

			if( ($backbone_1M2+$min_score_1M2) > 0 )
			{
				@{$rotamer_lib{$key1}{$key2}[$k]}=();
			#	$temp1=$backbone_1M2+$min_score_1M2;
			#	$delete_flag=$k;
			}
			elsif( ($backbone_2M1+$min_score_2M1) > 0 )
			{
				@{$rotamer_lib{$key1}{$key2}[$j]}=();
			#	$temp2=$backbone_2M1+$min_score_2M1;
			#	$delete_flag=$j;
			}
			#print "$key1\t$key2\tcompare $j $k\t$temp1 $temp2\tdelete $delete_flag\n";
		}
	}

	for($j=$iteration; $j>=0; $j--)
	{
		if($#{$rotamer_lib{$key1}{$key2}[$j]}==-1)
		{
			splice(@{$rotamer_lib{$key1}{$key2}},$j,1);
			splice(@{$score_single{$key1}{$key2}},$j,1);
		}
	}
	print "$key1\t$key2\t$#{$rotamer_lib{$key1}{$key2}}\n";
}


@current_combination=();
@idx_current=();
$max_score=-99999;
@max_combination=();
&exhaustive_search(\%side_chain_ref_chi, \%at_aryidx, \%Gate_Atom, \@sorted_sequence, \%rotamer_lib, \@current_combination, \@idx_current, \$max_score, \@max_combination);
print "max_score=$max_score\t combination => @max_combination\n";


### replace side chains of the gating residues by rotated ones
for($i=0; $i<=$#max_combination; $i++)
{
	$idx_comb=$max_combination[$i];
	$key1=$sorted_sequence[$i][0];
	$key2=$sorted_sequence[$i][1];
	$value=$sorted_sequence[$i][2];
	@rotamer=();
	&rotamer_generation($value, \@{$side_chain_ref_chi{$key1}{$key2}}, \%{$at_aryidx{$key1}{$key2}}, \@{$Gate_Atom{$key1}{$key2}}, \@{$rotamer_lib{$key1}{$key2}[$idx_comb]}, \@rotamer);
	# $pdb_obj->pdb_print("$key1\_$key2\_$value.pdb",\@rotamer);
	&replace_sidechain(\@ATOM,\@rotamer);
}
$output_filename="gate_reposition_".$pdb_file;
$pdb_obj->pdb_print($output_filename,\@ATOM);

($user1,$system1,$child1,$childsystem1)=times;
$user1-=$user;
$system1-=$system;
$child1-=$child;
$childsystem1-=$childsystem;
print "Step 3 : Dead End Elimination for the best combination of side chain rotamers : finished\n";
printf "running time = %.1f (s)\n\n", $user1;


### Total running time ###
($Tuser1,$Tsystem1,$Tchild1,$Tchildsystem1)=times;
$Tuser1-=$Tuser;
$Tsystem1-=$Tsystem;
$Tchild1-=$Tchild;
$Tchildsystem1-=$Tchildsystem;
printf "\nJob Completed!\nTotal running time = %.1f (s)\n\n", $Tuser1;
exit;



### sub routins start from here

### compute all dihedral angles of the side chain of a residue
### &side_chain_dihedral_angle($value,\@{$side_chain_ref_chi{$key1}{$key2}},\@{$Gate_Atom{$key1}{$key2}},\%{$at_aryidx{$key1}{$key2}});
sub side_chain_dihedral_angle
{
	my ($resn, $chi_aryref, $ATOM_aryref, $at_aryidx_hashref)=@_;
	my ($chi_number, $i);
	my (@temp1, @a0, @a1, @a2, @a3);

	$chi_number=$#{$Dihedral::definition{$resn}};
	for($i=0; $i<=$chi_number; $i++)
	{
		@temp1=();
		@temp1=@{$Dihedral::definition{$resn}[$i]};
		@a0=@a1=@a2=@a3=();
		@a0=@{$$ATOM_aryref[$$at_aryidx_hashref{$temp1[0]}]}[9..11];
		@a1=@{$$ATOM_aryref[$$at_aryidx_hashref{$temp1[1]}]}[9..11];
		@a2=@{$$ATOM_aryref[$$at_aryidx_hashref{$temp1[2]}]}[9..11];
		@a3=@{$$ATOM_aryref[$$at_aryidx_hashref{$temp1[3]}]}[9..11];
		$$chi_aryref[$i]=$vector_op->dihedral(\@a0,\@a1,\@a2,\@a3);
	}
}

### generate the coordinates of the side chain according to the rotamer library
### &rotamer_generation($value,\@{$side_chain_ref_chi{$key1}{$key2}},\%{$at_aryidx{$key1}{$key2}},\@{$Gate_Atom{$key1}{$key2}},\@temp,\@target_side_chain);
sub rotamer_generation
{
	my ($resn, $chi_aryref, $at_aryidx_hashref, $ATOM_aryref, $rotamer_dihedral_aryref, $target_residue_aryref)=@_;
	my ($i, $orig_CG1_idx, $orig_CG2_idx, $ILE_swap_flag, $rot_deg);
	my @temp;

	$vector_op->copy_MD_array($target_residue_aryref,$ATOM_aryref);
	$ILE_swap_flag=0;
	if($resn eq "ILE")
	{
		$orig_CG1_idx=$$at_aryidx_hashref{"CG1"};
		$orig_CG2_idx=$$at_aryidx_hashref{"CG2"};
		if( $orig_CG2_idx > $orig_CG1_idx )
		{
			$ILE_swap_flag=1;
			$$at_aryidx_hashref{"CG1"}=$orig_CG2_idx;
			$$at_aryidx_hashref{"CG2"}=$orig_CG1_idx;
			($$target_residue_aryref[$orig_CG1_idx],$$target_residue_aryref[$orig_CG2_idx])=($$target_residue_aryref[$orig_CG2_idx],$$target_residue_aryref[$orig_CG1_idx]);
		}
	}

	for($i=0; $i<=$#{$chi_aryref}; $i++)
	{
		@temp=();
		@temp=@{$Dihedral::definition{$resn}[$i]};
		$rot_deg=$$chi_aryref[$i]-$$rotamer_dihedral_aryref[$i];
		$quad_rot->side_chain_rotate(\@temp,$target_residue_aryref,$rot_deg);
	}

	if( $ILE_swap_flag==1 )
	{
		$$at_aryidx_hashref{"CG1"}=$orig_CG1_idx;
		$$at_aryidx_hashref{"CG2"}=$orig_CG2_idx;
		($$target_residue_aryref[$orig_CG1_idx],$$target_residue_aryref[$orig_CG2_idx])=($$target_residue_aryref[$orig_CG2_idx],$$target_residue_aryref[$orig_CG1_idx]);
	}
}


### compute the distance between the target atom and the closest atom from other residues
### ($clash_flag, $score_temp)=&serious_clash(\@target_side_chain, $tree, \@ATOM, \@gate_ca);
sub serious_clash
{
	my ($target_ATOM_aryref, $tree_ref, $ATOM_aryref, $gate_ca_aryref)=@_;
	my ($iteration, $i, $atm_idx, $min_r, $atom_type, $vdw_r, $score, $dis, $tolerance);
	my @SC_COM;

	$score=0;
	$iteration=$#{$target_ATOM_aryref};
	for($i=0; $i<=$iteration; $i++)
	{
		$atom_type=$$target_ATOM_aryref[$i][3];
		
		if($atom_type ne "N" && $atom_type ne "CA" && $atom_type ne "C" && $atom_type ne "O" && $atom_type ne "CB")
		{
			$vdw_r=$Parameter::vdw_radius{$$target_ATOM_aryref[$i][4]};

			($atm_idx,$min_r)=$kd_tree_op->nnsearch_kd_tree_r1($tree_ref,0,$ATOM_aryref,@{$$target_ATOM_aryref[$i]}[9..11]);
			$kd_tree_op->nnsearch_kd_tree_r2($tree_ref,0,\$atm_idx,\$min_r,$ATOM_aryref,@{$$target_ATOM_aryref[$i]}[9..11]);
			## if there is a serious clash than return, else calculate self score(potential) for the later DEE
			if($atom_type eq "CG" || $atom_type eq "CG1" || $atom_type eq "OG" || $atom_type eq "OG1" || $atom_type eq "SG")
			{
				$tolerance=$clash_tolerance_ga;
			}
			else
			{
				$tolerance=$clash_tolerance;
			}

			if(($vdw_r-$min_r) > $tolerance )
			{
				return (1,0); 
			}
			else
			{
				$score-=$min_r;
				if($number_of_gating_residue > 1)
				{
					@SC_COM=(); $dis=0;
					$SC_COM[0]=$$target_ATOM_aryref[$i][9]-$$gate_ca_aryref[0];
					$SC_COM[1]=$$target_ATOM_aryref[$i][10]-$$gate_ca_aryref[1];
					$SC_COM[2]=$$target_ATOM_aryref[$i][11]-$$gate_ca_aryref[2];
					$dis=$vector_op->norm_vec(@SC_COM);
					$score+=$dis;
				}
			}
			#print "$atom_type\t$vdw_r\t$atm_idx\t$min_r\n";
		}
	}
	return (0,$score);
}


### main subroutin for dead end elimination 
### ($min_score_1M2, $min_score_2M1)=&dead_end_elimination($i, \@sorted_sequence, \@rotamer_1, \@rotamer_2, \%rotamer_lib, \%side_chain_ref_chi, \%at_aryidx, \%Gate_Atom);
sub dead_end_elimination
{
	my ($target_idx, $sorted_sequence_aryref, $side_chain_1_aryref, $side_chain_2_aryref, $rotamer_lib_hashref, $side_chain_ref_chi_hashref, $at_aryidx_hashref, $Gate_Atom_hashref)=@_;
	my ($i, $j, $key1, $key2, $value, $iteration_1, $iteration_2, $chain_id, $resi, $temp_score_1, $temp_score_2, $total_score_1M2, $total_score_2M1, $min_1M2, $min_2M1, $tmp_1M2, $tmp_2M1);
	my @rotamer_con=();

	$total_score_1M2=$total_score_2M1=0;
	$chain_id=$$sorted_sequence_aryref[$target_idx][0];
	$resi=$$sorted_sequence_aryref[$target_idx][1];
	$iteration_1=$#{$sorted_sequence_aryref};
	for($i=0; $i<=$iteration_1; $i++)
	{
		if($i==$target_idx)
		{
			next;
		}

		$key1=$$sorted_sequence_aryref[$i][0];
		$key2=$$sorted_sequence_aryref[$i][1];
		$value=$$sorted_sequence_aryref[$i][2];
		$iteration_2=$#{$$rotamer_lib_hashref{$key1}{$key2}};

		$min_1M2=$min_2M1=99999;
		for($j=0; $j<=$iteration_2; $j++)
		{
			@rotamer_con=();
			&rotamer_generation($value,\@{$$side_chain_ref_chi_hashref{$key1}{$key2}},\%{$$at_aryidx_hashref{$key1}{$key2}},\@{$$Gate_Atom_hashref{$key1}{$key2}},\@{$$rotamer_lib_hashref{$key1}{$key2}[$j]},\@rotamer_con);
			# $pdb_obj->pdb_print("$key1\_$key2\_$j.pdb",\@rotamer_con);
		
			$temp_score_1=&interaction_score($side_chain_1_aryref, \@rotamer_con);
			$temp_score_2=&interaction_score($side_chain_2_aryref, \@rotamer_con);
			#print "$temp_score_1\t$temp_score_2\n";
	
			$tmp_1M2=$temp_score_1-$temp_score_2;
			$tmp_2M1=-1*$tmp_1M2;
			$min_1M2=$tmp_1M2 < $min_1M2 ? $tmp_1M2:$min_1M2;
			$min_2M1=$tmp_2M1 < $min_2M1 ? $tmp_2M1:$min_2M1;
			#print "min 1M2=$min_1M2\t2M1=$min_2M1\n";
		}
		$total_score_1M2+=$min_1M2;
		$total_score_2M1+=$min_2M1;
	}
	return($total_score_1M2, $total_score_2M1);
}


### computation of interaction score between side chains of two residues
### $temp_score_1=&interaction_score($side_chain_1_aryref, \@rotamer_con);
sub interaction_score
{
	my ($rotamer_aryref, $target_rotamer_aryref)=@_;
	my ($length, $length_target, $i, $j, $k, $score, $atm_type_1, $atm_type_2, $dis_1, $vdw_r1, $vdw_r2);
	my @v1;

	$length=$#{$rotamer_aryref};
	$length_target=$#{$target_rotamer_aryref};
	$score=0;
	for($i=0; $i<=$length; $i++)
	{
		$atm_type_1=$$rotamer_aryref[$i][3];
		if($atm_type_1 eq "N" || $atm_type_1 eq "C" || $atm_type_1 eq "CA" || $atm_type_1 eq "O" )
		{
			next;
		}
		$vdw_r1=$Parameter::vdw_radius{$$rotamer_aryref[$i][4]};
		
		for($j=0; $j<=$length_target; $j++)
		{
			$atm_type_2=$$target_rotamer_aryref[$j][3];
			if($atm_type_2 eq "N" || $atm_type_2 eq "C" || $atm_type_2 eq "CA" || $atm_type_2 eq "O" )
			{
				next;
			}
			$vdw_r2=$Parameter::vdw_radius{$$target_rotamer_aryref[$j][4]};

			@v1=();
			for($k=0; $k<=2; $k++)
			{
				$v1[$k]=$$rotamer_aryref[$i][9+$k]-$$target_rotamer_aryref[$j][9+$k];
			}
			$dis_1=$vector_op->norm_vec(@v1)-$vdw_r1-$vdw_r2;
			$score+=$dis_1;
		}
	}
	return($score);
}

### exhaustive search of the combinations 
### &exhaustive_search(\%side_chain_ref_chi, \%at_aryidx, \%Gate_Atom, \@sorted_sequence, \%rotamer_lib, \@current_combination, \@idx_current, \$max_score, \@max_combination);
sub exhaustive_search
{
	my ($side_chain_ref_chi_hashref, $at_aryidx_hashref, $Gate_Atom_hashref, $sorted_sequence_aryref, $rotamer_lib_hashref, $rotamer_combination_aryref, $idx_combination, $max_score_ref, $max_combination_aryref)=@_;
	my ($total_level, $level, $key1, $key2, $value, $iteration, $i, $j, $k, $score);
	my (@current_combination, @single_rotamer, @idx_current);

	$total_level=$#{$sorted_sequence_aryref};
	$level=$#{$idx_combination}+1;
	$key1=$$sorted_sequence_aryref[$level][0];
	$key2=$$sorted_sequence_aryref[$level][1];
	$value=$$sorted_sequence_aryref[$level][2];
	$iteration=$#{$$rotamer_lib_hashref{$key1}{$key2}};

	print "in subroutin level=$level\t@{$idx_combination}\n";
	print "$key1\t$key2\t$value\n";

	for( $i=0; $i<=$iteration; $i++ )
	{
		@idx_current=();
		@idx_current=@{$idx_combination};
		push(@idx_current,$i);
		
		@current_combination=();
		$vector_op->copy_MD_array(\@current_combination,$rotamer_combination_aryref);
		@single_rotamer=();
		&rotamer_generation($value, \@{$$side_chain_ref_chi_hashref{$key1}{$key2}}, \%{$$at_aryidx_hashref{$key1}{$key2}}, \@{$$Gate_Atom_hashref{$key1}{$key2}}, \@{$$rotamer_lib_hashref{$key1}{$key2}[$i]}, \@single_rotamer);
		$vector_op->copy_MD_array(\@{$current_combination[$level]}, \@single_rotamer);
		

		if($level==$total_level)
		{
			$score=0;
			for($j=0; $j<=$level-1; $j++)
			{
				for($k=$j+1; $k<=$level; $k++)
				{
					$score+=&interaction_score(\@{$current_combination[$j]}, \@{$current_combination[$k]});
				}
			}

			for($j=0; $j<=$level; $j++)
			{
				$key1_here=$current_combination[$j][0][7];
				$key2_here=$current_combination[$j][0][8];
				$idx_here=$idx_current[$j];
				print "$current_combination[$j][0][0]\t$key1_here\t$key2_here\t$idx_here\n";
				$score+=$score_single{$key1_here}{$key2_here}[$idx_here];
			}
			
			print"score=$score\n";
			if($score > $$max_score_ref)
			{
				$$max_score_ref=$score;
				@{$max_combination_aryref}=();
				@{$max_combination_aryref}=@idx_current;
				print "max_score_ref=$$max_score_ref\n";
			}
		}
		elsif($level < $total_level)
		{
			&exhaustive_search($side_chain_ref_chi_hashref, $at_aryidx_hashref, $Gate_Atom_hashref, $sorted_sequence_aryref, $rotamer_lib_hashref, \@current_combination, \@idx_current, $max_score_ref, $max_combination_aryref);
		}
	}
}


### replace side chain of the gating residue
### &replace_sidechain(\@ATOM,\@rotamer);
sub replace_sidechain
{
	my ($ATOM_ref, $rotamer_ref)=@_;
	my ($ATOM_iter, $rotamer_iter, $i, $start_idx);

	$ATOM_iter=$#{$ATOM_ref};
	$rotamer_iter=$#{$rotamer_ref};
	
	for($i=0; $i<=$ATOM_iter; $i++)
	{
		if($$ATOM_ref[$i][6] eq $$rotamer_ref[0][6] && $$ATOM_ref[$i][7] eq $$rotamer_ref[0][7] && $$ATOM_ref[$i][8] eq $$rotamer_ref[0][8])
		{
			$start_idx=$i;
			last;
		}
	}

	for($i=0; $i<=$rotamer_iter; $i++)
	{
		$$ATOM_ref[$i+$start_idx][9]=$$rotamer_ref[$i][9];
		$$ATOM_ref[$i+$start_idx][10]=$$rotamer_ref[$i][10];
		$$ATOM_ref[$i+$start_idx][11]=$$rotamer_ref[$i][11];
	}
}


### get and check the input parameters
sub get_check_options
{
	my %opts=();
	my ($pdb_file, $gate_list, $clash_tolerance)=("", "", 1.0); # defaults values

	getopt('fith',\%opts);

	if( exists $opts{'h'} )
	{
		&help_test();
		exit;
	}

	if( !exists $opts{'f'} )
	{
		print "\nPlease input pdb file of protein\n";
		&help_test();
		exit;
	}
	else
	{
		$pdb_file=$opts{'f'};
	}

	if( !exists $opts{'i'} )
	{
		print "\nPlease input the list of gating residues\n";
		&help_test();
		exit;
	}
	else
	{
		$gate_list=$opts{'i'};
	}

	if( exists $opts{'t'} )
	{
		$clash_tolerance=$opts{'t'};
	}



	### check for some mistakes ###
	if($pdb_file eq "" || $gate_list eq "")
	{
		print "\nPlease input the pdb file of protein and the list of gating residues\n";
		&help_test();
		exit;
	}

	if($clash_tolerance < 0)
	{
		print "\nThe tolerance of steric clashes must be a positive value\n";
		&help_test();
		exit;
	}


	print "\nRunning perl Gate_Open.pl -f $pdb_file -i $gate_list -t $clash_tolerance\n\n";
	return($pdb_file, $gate_list, $clash_tolerance);
}


### usage of this program
sub help_test
{
	print "\nusage : perl Gate_Open.pl -f pdb_file -i gating_residue_list -t clash_tolerance\n";
	print "\t-f\tPDB file of the targeted protein\n";
	print "\t-i\tfile contains the list of gating residues\n";
	print "\t-t\tThe tolerance of steric clashes (default: 1.0 angstrom)\n";
	print "\t-h\tshow this page\n\n";
}

