############################################################################
#                                                                          #
#   Copyright (C) 2011 Po-Hsien Lee <p.lee@mx.uni-saarland.de>             #
#   All rights reserved.                                                   #
#                                                                          #
#   This file is part of PROPORES.                                         #
#                                                                          #
#   PROPORES is free software: you can redistribute it and/or modify       #
#   it under the terms of the GNU General Public License as published by   #
#   the Free Software Foundation, either version 3 of the License, or      #
#   (at your option) any later version.                                    #
#                                                                          #
#   PROPORES is distributed in the hope that it will be useful,            #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of         #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
#   GNU General Public License for more details.                           #
#                                                                          #
#   You should have received a copy of the GNU General Public License      #
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.  #
#                                                                          #
############################################################################
#                                                                          #
#   Pore_Trace.pl                                                          #
#                                                                          #
#   Programmed by Po-Hsien Lee <p.lee@mx.uni-saarland.de>, 2011            #
#   Center for Bioinformatics, Saarland University, Germany                #
#                                                                          #
#   Pore_Trace.pl is a program for determining pore axes of the pores      #
#   identified by Pore_ID.pl.                                              #
#                                                                          #
############################################################################

#!/usr/bin/perl

### include necessary perl module and initialization
use POSIX;
use Vector;
use Getopt::Std;

$vector_op=Vector->new();
$perturb_value=0.01;
($Tuser,$Tsystem,$Tchild,$Tchildsystem)=times;


### get and check input parameters
($pore_PTin_file)=&get_check_options();

if($pore_PTin_file=~m/(\S+)\.PTin/)
{
	$output_prefix=$1;
}


### read in the input file
@ATOM=(); $count=0;
open FH, "<$pore_PTin_file" or die "$!";
while(<FH>)
{
	if(m/^#/)
	{
		if(m/Grid side length = (\S+) \(angstrom\)/)
		{
			$grid_side_length=$1;
		}
	}
	else
	{
		@temp=();
		@temp=split(/\s+/,$_);
		@{$ATOM[$count]}=@temp;
		$count++;
	}
}
close FH;


### define the boundary of the void grid voxels
$x_upper=$y_upper=$z_upper=-99999;
$x_lower=$y_lower=$z_lower=99999;
for($i=0;$i<=$#ATOM;$i++)
{
        &pdb_boundary(\$x_upper,\$x_lower,$ATOM[$i][0]);
        &pdb_boundary(\$y_upper,\$y_lower,$ATOM[$i][1]);
        &pdb_boundary(\$z_upper,\$z_lower,$ATOM[$i][2]);
}
($endx, $endy, $endz)=&gridize($x_upper, $y_upper, $z_upper);


### initialization of matrix recorded the scoring
@pore_grid=(); @block_or_not=(); @score_array=();
for($i=0; $i<=$#ATOM; $i++)
{
	($x,$y,$z)=&gridize($ATOM[$i][0],$ATOM[$i][1],$ATOM[$i][2]);
	$pore_grid[$i][0]=$x;
	$pore_grid[$i][1]=$y;
	$pore_grid[$i][2]=$z;
	$pore_grid[$i][3]=99999; # current score
	$block_or_not[$x][$y][$z][0]=1;  # 1 for exist, 2 for blocked
	$block_or_not[$x][$y][$z][1]=$i; # index of score array of current score
	$block_or_not[$x][$y][$z][2]=0;  # length of path
	$block_or_not[$x][$y][$z][3]=0;  # current score
	$block_or_not[$x][$y][$z][4]=1/($ATOM[$i][3]+$perturb_value); # self score for 1/R_ball
	$block_or_not[$x][$y][$z][5]=($ATOM[$i][5]+$perturb_value)/($ATOM[$i][3]+$perturb_value);  # self score for D_com/R_ball
	$block_or_not[$x][$y][$z][6]=$ATOM[$i][4]; # exposed pore-surface is 2 , non-exposed-surface is 1
	$block_or_not[$x][$y][$z][10]=$ATOM[$i][3];
}



### determine the grid position (surface of inside)
#open FH, "> exposed.pdb " or die "$!"; $temp=0;
#open FH2, "> not_exposed.pdb" or die "$!"; $temp2=0;
#for($i=0; $i<=$endx; $i++)
#{
#	for($j=0; $j<=$endy; $j++)
#	{
#		for($k=0; $k<=$endz; $k++)
#		{
#			if($block_or_not[$i][$j][$k][6]==2)
#			{
#				$temp++;
#				($coor_i,$coor_j,$coor_k)=&inverse($i,$j,$k);
#				&print_pseduPDB(FH,$temp,1,$coor_i,$coor_j,$coor_k);
#			}
#			elsif($block_or_not[$i][$j][$k][6]==1)
#			{
#				$temp2++;
#				($coor_i,$coor_j,$coor_k)=&inverse($i,$j,$k);
#				&print_pseduPDB(FH2,$temp,1,$coor_i,$coor_j,$coor_k);
#			}
#		}
#	}
#}
#close FH; close FH2;


@expose_cluster=(); @expose_info=();
&cluster(\@block_or_not, \@expose_cluster, \@expose_info);
@expose_cluster=();

$expose_info[0][1]=0;
@expose_info=sort{ $a->[1] <=> $b->[1] }(@expose_info);
@expose_info=reverse(@expose_info);


############################# skip a small surface patch
$exposed_size_criteria=30; # set as a larger value if you want to skip a small surface patch
for($i=$#expose_info; $i>=0; $i--)
{
	if( $expose_info[$i][1] < $exposed_size_criteria )
	{
		for($j=0; $j<=$#{$expose_info[$i][2]}; $j++)
		{
			$tmpx=$expose_info[$i][2][$j][0];
			$tmpy=$expose_info[$i][2][$j][1];
			$tmpz=$expose_info[$i][2][$j][2];
			$block_or_not[$tmpx][$tmpy][$tmpz][6]=1;
		}
		splice(@expose_info, $i, 1)
	}
}
#############################


### figure out the endpoints of each surface patches
@start_point=();
for($i=0; $i<=$#expose_info; $i++)
{
		$min_self_score=99999; @min_self_grid=();
		$iter=$expose_info[$i][1]-1;
		for($j=0;$j<=$iter;$j++)
		{
			$tmpx=$expose_info[$i][2][$j][0];
			$tmpy=$expose_info[$i][2][$j][1];
			$tmpz=$expose_info[$i][2][$j][2];
			$tmp_self_score=$block_or_not[$tmpx][$tmpy][$tmpz][4];
			if($tmp_self_score < $min_self_score)
			{
				$min_self_score=$tmp_self_score;
				@min_self_grid=($tmpx,$tmpy,$tmpz);
			}
		}
		open FH, ">$output_prefix\_endpoint_$i.pdb" or die "$!";
		($coor_i,$coor_j,$coor_k)=&inverse(@min_self_grid);
		&print_pseduPDB(FH,$temp,1,$coor_i,$coor_j,$coor_k);
		close FH;
		@{$start_point[$i]}=@min_self_grid;
}
$patch_number=$#start_point+1;
print "Number of surface patches (endpoints): $patch_number\n";


if( $#start_point>=1 )
{
	$tag="TWO";
}
elsif( $#start_point==0 )
{
	$tag="ONE";
}
elsif( $#start_point < 0 )
{
	$tag="ZERO";
}

### if no expose surface then find the farest point from center as start point
if( $tag eq "ZERO" )
{
	@COM=(); $max_d_sq=-99999;
	$COM[0]=($x_upper+$x_lower)/2;
	$COM[1]=($y_upper+$y_lower)/2;
	$COM[2]=($z_upper+$z_lower)/2;

	for($i=0; $i<=$#ATOM; $i++)
	{
		if($ATOM[$i][4]==1)
		{
			@temp_vec=();
			$temp_vec[0]=$COM[0]-$ATOM[$i][0];
			$temp_vec[1]=$COM[1]-$ATOM[$i][1];
			$temp_vec[2]=$COM[2]-$ATOM[$i][2];
			$dis_sq=$vector_op->norm_vec_square(@temp_vec);
	
			if($dis_sq > $max_d_sq)
			{
				$max_d_sq=$dis_sq;
				($start_point[0][0], $start_point[0][1], $start_point[0][2])=&gridize($ATOM[$i][0], $ATOM[$i][1], $ATOM[$i][2]);
			}
		}
	}
}


### Dijkstra's algorithm to find the loest cost path as the pore axis
if( $tag eq "ZERO" || $tag eq "ONE")
{
	$iter=0;
}
elsif( $tag eq "TWO" )
{
	$iter=$#start_point-1;
}

@search_seq_1=();
&search_sequence(1, \@search_seq_1);

for($i=0; $i<=$iter; $i++)
{
	&Dijkstra($i, \@start_point, \@block_or_not, \@pore_grid, $tag, \@search_seq_1);
}



### Total running time 
($Tuser1,$Tsystem1,$Tchild1,$Tchildsystem1)=times;
$Tuser1-=$Tuser;
$Tsystem1-=$Tsystem;
$Tchild1-=$Tchild;
$Tchildsystem1-=$Tchildsystem;
printf "\nJob Completed!\nTotal running time = %.1f (s)\n\n", $Tuser1;

exit;




### rests are subroutins

### Dijkstra's algorithm for searching lowest-cost path
# &Dijkstra(1, \@start_point, \@block_or_not, \@pore_grid, $tag, \@search_seq_1);
sub Dijkstra
{

	my ($start_idx, $start_point_aryref, $block_or_not_aryref, $pore_grid_aryref, $tag, $search_seq_1_aryref)=@_;
	my ($x, $y, $z, $idx_tmp, $i, $j, $k, $m, $index1, $index2, $idx_pre, $previous_score, $previous_length, $current_self_score, $score_new, $idx_now, $old_score, $temp, $coor_i, $coor_j, $coor_k, $x_old, $y_old, $z_old, $x_new, $y_new, $z_new, $min_score, $tag_index, $lowest_score_index, $discard_index);
	my (@block_or_not_buf, @pore_grid_buf, @current_affect_element, %hash_in_affect);
	

	# block_or_not sould be copy for buffer
	# pore_grid must copy for buffer
	@block_or_not_buf=@pore_grid_buf=();
	$vector_op->copy_MD_array(\@block_or_not_buf, $block_or_not_aryref);
	$vector_op->copy_MD_array(\@pore_grid_buf, $pore_grid_aryref);


	# different scoring scheme
	$tag_index=0;
	if($tag eq "ZERO" || $tag eq "TWO")
	{
		$tag_index=0;
	}
	elsif($tag eq "ONE")
	{
		$tag_index=1;
	}

	### determine source grid point
	($x, $y, $z)=(@{$$start_point_aryref[$start_idx]});
	$idx_tmp=$block_or_not_buf[$x][$y][$z][1];
	$pore_grid_buf[$idx_tmp][3]=$block_or_not_buf[$x][$y][$z][4+$tag_index];
	$block_or_not_buf[$x][$y][$z][2]=1; # length of path
	$block_or_not_buf[$x][$y][$z][7]=$x; # traceback x
	$block_or_not_buf[$x][$y][$z][8]=$y; # traceback y
	$block_or_not_buf[$x][$y][$z][9]=$z; # traceback z
	


	@current_affect_element=(); %hash_in_affect=();
	push(@current_affect_element,$idx_tmp);
	$hash_in_affect{$idx_tmp}=1;
	$remaining_element=$#pore_grid_buf;

	### forward iteration part
	while(  $#current_affect_element >= 0 )
	{

		$lowest_score_index=$current_affect_element[0];
		$discard_index=0;
		$lowest_score=$pore_grid_buf[$lowest_score_index][3];
		## find a grid point in the rest portion with minimum score 
		for($i=1; $i<=$#current_affect_element; $i++)
		{
			$index1=$current_affect_element[$i];
			$temp_score=$pore_grid_buf[$index1][3];
			if( $temp_score < $lowest_score )
			{
				$lowest_score_index=$index1;
				$lowest_score=$temp_score;
				$discard_index=$i;
			}
		}

		
		if($pore_grid_buf[$lowest_score_index][3]==99999)
		{
			print "something wrong in Dijkstra's algorithm\n";
			last;
		}

		$x=$pore_grid_buf[$lowest_score_index][0];
		$y=$pore_grid_buf[$lowest_score_index][1];
		$z=$pore_grid_buf[$lowest_score_index][2];
		#print "$x\t$y\t$z\t$pore_grid_buf[$lowest_score_index][3]\n"; 
		
		$block_or_not_buf[$x][$y][$z][0]=2; # block current grid
		for($m=1; $m<=$#$search_seq_1_aryref; $m++)
		{
			$i=$x+$$search_seq_1_aryref[$m][1];
			$j=$y+$$search_seq_1_aryref[$m][2];
			$k=$z+$$search_seq_1_aryref[$m][3];
			if( $block_or_not_buf[$i][$j][$k][0]==1 && $i>=0 && $j>=0 && $k>=0 )
			{
				$idx_pre=$block_or_not_buf[$x][$y][$z][1];
				$previous_score=$pore_grid_buf[$idx_pre][3];
				$previous_length=$block_or_not_buf[$x][$y][$z][2];
				$current_self_score=$block_or_not_buf[$i][$j][$k][4+$tag_index]; # 4 or 5
				$score_new=($previous_score+$current_self_score);
				$idx_now=$block_or_not_buf[$i][$j][$k][1];
				$old_score=$pore_grid_buf[$idx_now][3];
				if($score_new < $old_score)
				{
					$pore_grid_buf[$idx_now][3]=$score_new;
					$block_or_not_buf[$i][$j][$k][2]=$previous_length+1;
					$block_or_not_buf[$i][$j][$k][7]=$x;
					$block_or_not_buf[$i][$j][$k][8]=$y;
					$block_or_not_buf[$i][$j][$k][9]=$z;
				}
				
				if($hash_in_affect{$idx_now}!=1)
				{
					$hash_in_affect{$idx_now}=1;
					push(@current_affect_element,$idx_now);
				}
			}
		}
		$block_or_not_buf[$x][$y][$z][0]=2; # block previous grid
		splice(@current_affect_element,$discard_index,1);

		$remaining_element--;
		#print "length of pore_grid_buf=$remaining_element\n";
	}


	### traceback iteration part
	if( $tag eq "ZERO" || $tag eq "ONE")
	{
		$min_score=99999;
		for($i=0; $i<=$endx; $i++)
		{
			for($j=0; $j<=$endy; $j++)
			{
				for($k=0; $k<=$endz; $k++)
				{
					if($block_or_not_buf[$i][$j][$k][6]==1 && $block_or_not_buf[$i][$j][$k][2]!=1 )
					{
						$idx_now=$block_or_not_buf[$i][$j][$k][1];
						$score_new=$pore_grid_buf[$idx_now][3]/($block_or_not_buf[$i][$j][$k][2]**1.5); #total divided by length
						if($score_new < $min_score)
						{
							$min_score=$score_new;
							@{$$start_point_aryref[1]}=($i, $j, $k);
						}
					}
				}
			}
		}
	}

	
	if($tag eq "ONE")
	{
		&Dijkstra(0, $start_point_aryref, $block_or_not_aryref, $pore_grid_aryref, "TWO", $search_seq_1_aryref);
		return;
	}

	


	for($i=0; $i<=$#{$start_point_aryref}; $i++)
	{
		if($i<=$start_idx)
		{
			next;
		}
	
		open FH, "> $output_prefix\_axis_$start_idx\_to_$i.pdb" or die "$!"; $temp=0;
		($x_old, $y_old, $z_old)=(@{$$start_point_aryref[$i]});
		$idx_tmp=$block_or_not_buf[$x_old][$y_old][$z_old][1];
		#print FH "REMARK score=$score_array_buf[$idx_tmp]\n";
		$temp++;
		($coor_i,$coor_j,$coor_k)=&inverse($x_old, $y_old, $z_old);
		&print_pseduPDB(FH,$temp,1,$coor_i,$coor_j,$coor_k, $block_or_not_buf[$x_old][$y_old][$z_old][10]);

		do
		{
			($x_new, $y_new, $z_new)=($block_or_not_buf[$x_old][$y_old][$z_old][7],$block_or_not_buf[$x_old][$y_old][$z_old][8],$block_or_not_buf[$x_old][$y_old][$z_old][9]);
			$temp++;
			($coor_i, $coor_j, $coor_k)=&inverse($x_new, $y_new, $z_new);
			&print_pseduPDB(FH, $temp, 1, $coor_i, $coor_j, $coor_k, $block_or_not_buf[$x_new][$y_new][$z_new][10]);
			
			$x_old=$x_new;
			$y_old=$y_new;
			$z_old=$z_new;
		}while($x_new!=$block_or_not_buf[$x_new][$y_new][$z_new][7] || $y_new!=$block_or_not_buf[$x_new][$y_new][$z_new][8] || $z_new!=$block_or_not_buf[$x_new][$y_new][$z_new][9]);
		close FH;
	}
}


### define the extreme value as the boundary
sub pdb_boundary
{
	my ($upper_ref, $lower_ref, $target)=@_;

	if($target > $$upper_ref)
	{
		$$upper_ref=$target;
	}
	if($target < $$lower_ref)
	{
		$$lower_ref=$target;
	}
}


### gridize a 3D coordinate
### &gridize($x,$y,$z);
sub gridize
{
	my $i;
	my (@grid_output, @grid_tmp, @lower);
	@lower=($x_lower,$y_lower,$z_lower);

	for($i=0; $i<=2; $i++)
	{
		$grid_output[$i]=int(($_[$i]-$lower[$i])/$grid_side_length);
		$grid_tmp[$i]=($_[$i]-$lower[$i])/$grid_side_length;
		if(($grid_tmp[$i]-$grid_output[$i])>=0.5)
		{
			$grid_output[$i]++;
		}
	}
	return (@grid_output);
}


### print a point in PDB format
sub print_pseduPDB
{
	my ($file_handle, $index, $resi, $coor_i, $coor_j, $coor_k, $radius)=@_;

	if($index<=99999)
	{
		if($#_ == 5)
		{
			printf $file_handle "ATOM  %5d  C   AAA   %3d    %8.3f%8.3f%8.3f\n",$index,$resi,$coor_i,$coor_j,$coor_k;
		}
		elsif($#_ == 6)
		{
			printf $file_handle "ATOM  %5d  C   AAA   %3d    %8.3f%8.3f%8.3f\t%.3f\n", $index, $resi, $coor_i, $coor_j, $coor_k, $radius;
		}
	}
	else
	{
		if($#_ == 5)
		{
			printf $file_handle "ATOM  *****  C   AAA   %3d    %8.3f%8.3f%8.3f\n",$resi,$coor_i,$coor_j,$coor_k;
		}
		elsif($#_ == 6)
		{
			printf $file_handle "ATOM  *****  C   AAA   %3d    %8.3f%8.3f%8.3f\t%.3f\n", $resi, $coor_i, $coor_j, $coor_k, $radius;
		}
	}
}


### inversion of gridize
sub inverse
{
	my ($coor_i, $coor_j, $coor_k)=(0.0, 0.0, 0.0);

	$coor_i=$_[0]*$grid_side_length+$x_lower;
	$coor_j=$_[1]*$grid_side_length+$y_lower;
	$coor_k=$_[2]*$grid_side_length+$z_lower;
	return($coor_i,$coor_j,$coor_k);
}


### cluster void grid voxels to form a surface patch
### &cluster(\@block_or_not,\@expose_cluster,\@expose_info);
sub cluster
{
	my ($I, $J, $K, $total); my $cluster_index=0;
	my ($block_or_not_aryref, $expose_cluster_aryref, $expose_info_aryref)=@_;
	my @search_seq_1=();
	
	&search_sequence(1, \@search_seq_1);
	
	for($I=0; $I<=$endx; $I++)
	{
		for($J=0; $J<=$endy; $J++)
		{
			for($K=0; $K<=$endz; $K++)
			{
				if( $$block_or_not_aryref[$I][$J][$K][6]==2 && $$expose_cluster_aryref[$I][$J][$K]<1 )
				{
					$total=0;
					$cluster_index++;
					$total=&neighbor_join($I, $J, $K, $block_or_not_aryref, $expose_cluster_aryref, $expose_info_aryref, $cluster_index, \@search_seq_1);
					$$expose_info_aryref[$cluster_index][0]=$cluster_index;
					$$expose_info_aryref[$cluster_index][1]=$total;
				}
			}
		}
	}
}


### main subroutin of cluster
sub neighbor_join
{
	my ($x, $y, $z, $block_or_not_aryref, $expose_cluster_aryref, $expose_info_aryref, $index, $search_seq_1_aryref)=@_;
	my ($sum, $idx_temp, $i, $j, $k, $m);


	#print "$x $y $z $index\n";
	$$expose_cluster_aryref[$x][$y][$z]=$index;

	$$expose_info_aryref[$index][0]=$index;
	$idx_temp=$#{$$expose_info_aryref[$index][2]};
	$idx_temp++;
	$$expose_info_aryref[$index][2][$idx_temp][0]=$x;
	$$expose_info_aryref[$index][2][$idx_temp][1]=$y;
	$$expose_info_aryref[$index][2][$idx_temp][2]=$z;

	$sum=1;
	for($m=1; $m<=$#$search_seq_1_aryref; $m++)
	{
		$i=$x+$$search_seq_1_aryref[$m][1];
		$j=$y+$$search_seq_1_aryref[$m][2];
		$k=$z+$$search_seq_1_aryref[$m][3];
		
		if($i<0 || $i>$endx || $j<0 || $j>$endy || $k<0 || $k>$endz)
		{
			next;
		}
		
		if( $$block_or_not_aryref[$i][$j][$k][6]==2 && $$expose_cluster_aryref[$i][$j][$k]<1 )
		{
			$sum+=&neighbor_join($i, $j, $k, $block_or_not_aryref, $expose_cluster_aryref, $expose_info_aryref, $index, $search_seq_1_aryref);
		}
	}

	return($sum);
}


### search sequence for effeciently search in the region of the grid
### &search_sequence($level, \@search_seq);
sub search_sequence
{
	my ($level,$search_seq_aryref)=@_;
	my ($count, $i, $j, $k);
	my @vec;

	$count=0;
	for($i=-$level; $i<=$level; $i++)
	{
		for($j=-$level; $j<=$level; $j++)
		{
			for($k=-$level; $k<=$level; $k++)
			{
				@vec=();
				@vec=($i,$j,$k);
				$$search_seq_aryref[$count][0]=$vector_op->norm_vec_square(@vec);
				$$search_seq_aryref[$count][1]=$i;
				$$search_seq_aryref[$count][2]=$j;
				$$search_seq_aryref[$count][3]=$k;
				$count++;
			}
		}
	}
	@{$search_seq_aryref}=sort{$a->[0] <=> $b->[0]} (@{$search_seq_aryref});
}


### get and check input parameter
sub get_check_options
{
	my %opts=();
	my $pore_PTin_file=""; # default value

	getopt('fh',\%opts);

	if( exists $opts{'h'} )
	{
		&help_test();
		exit;
	}
	
	if( !exists $opts{'f'} )
	{
		print "\nPlease input the pore_#.PTin file ( the output of Pore_ID calculation )\n";
		&help_test();
		exit;
	}
	else
	{
		$pore_PTin_file=$opts{'f'};
	}


	### check for some mistakes ###
	if( $pore_PTin_file eq "")
	{
		print "\nPlease input the pore_#.PTin file ( the output of Pore_ID calculation )\n";
		&help_test();
		exit;
	}


	print "\nRunning perl Pore_Trace.pl -f $pore_PTin_file\n\n";
	return($pore_PTin_file);
}


### usage of this program
sub help_test
{
	print "\nusage : perl Pore_Trace.pl -f pore_#.PTin\n";
	print "\t-f\tpore_#.PTin file (the output from Pore_ID calculation)\n";
	print "\t-h\tshow this page\n\n";
}
