############################################################################
#                                                                          #
#   Copyright (C) 2011 Po-Hsien Lee <p.lee@mx.uni-saarland.de>             #
#   All rights reserved.                                                   #
#                                                                          #
#   This file is part of PROPORES.                                         #
#                                                                          #
#   PROPORES is free software: you can redistribute it and/or modify       #
#   it under the terms of the GNU General Public License as published by   #
#   the Free Software Foundation, either version 3 of the License, or      #
#   (at your option) any later version.                                    #
#                                                                          #
#   PROPORES is distributed in the hope that it will be useful,            #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of         #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
#   GNU General Public License for more details.                           #
#                                                                          #
#   You should have received a copy of the GNU General Public License      #
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.  #
#                                                                          #
############################################################################
#!/usr/bin/perl


package PDBreader;

use Parameter;

sub new
{
	my $type = $_[0];
	my $self = {};
	bless($self, $type);
	return($self);
}


### read in a PDB file and store all columns in a multidimensional array
sub pdbparser
{
	my $file_name=$_[1]; my $array_ref=$_[2]; my $deHETATM=$_[3];

	my $count=0;
	open FH, "<$file_name" or die "Cannot open file $file_name\n$!";
	while(<FH>)
	{
		if(m/^ATOM/ || m/^HETATM/)
		{
			if(m/^HETATM/ && $deHETATM eq "deHETATM")
			{
				next;
			}

			$$array_ref[$count][0]=substr($_,0,6);		# atom or heteroatom
			$$array_ref[$count][1]=substr($_,6,5)+0;	# atom index
			$$array_ref[$count][2]=substr($_,12,4);		# atom type full original
			$$array_ref[$count][3]=$$array_ref[$count][2];
			$$array_ref[$count][3]=~s/\s+//g;		# atom type full chomped
			$$array_ref[$count][4]=substr($_,13,1);		# atom type single
			$$array_ref[$count][5]=substr($_,16,1);		# alternative position index
			$$array_ref[$count][6]=substr($_,17,3);		# residue type
			$$array_ref[$count][7]=substr($_,21,1);		# chain index
			$$array_ref[$count][8]=substr($_,22,4)+0;	# residue index
			$$array_ref[$count][9]=substr($_,30,8)+0.0;	# x coordinate
			$$array_ref[$count][10]=substr($_,38,8)+0.0;	# y coordinate
			$$array_ref[$count][11]=substr($_,46,8)+0.0;	# z coordinate
			$$array_ref[$count][12]=substr($_,54,6)+0.0;	# occupancy
			$$array_ref[$count][13]=substr($_,60,6)+0.0;	# temperature factor
			$count++;
		}
	}
}


### print out PDB format of the multidimensional array stored PDB structure
sub pdb_print
{
	my $output_name=$_[1]; my $array_ref=$_[2];
	my $length=$#$array_ref;
	my $FH;

	open $FH, ">$output_name" or die "$!";
	for(my $i=0; $i<=$length; $i++)
	{
		printf $FH "%6s%5d %4s%1s%3s %1s%4d    %8.3f%8.3f%8.3f%6.2f%6.2f\n",$$array_ref[$i][0],$$array_ref[$i][1],$$array_ref[$i][2],$$array_ref[$i][5],$$array_ref[$i][6],$$array_ref[$i][7],$$array_ref[$i][8],$$array_ref[$i][9],$$array_ref[$i][10],$$array_ref[$i][11],$$array_ref[$i][12],$$array_ref[$i][13];
	}
	close $FH;
}


### compute the center of mass of a protein
sub pdb_com
{
	my $ATOM_aryref=$_[1];
	my $heavy_only=$_[2];
	my ($i, $x_sum, $y_sum, $z_sum, $mass_sum, $mass_of, $com_x, $com_y, $com_z);
	$x_sum=$y_sum=$z_sum=$mass_sum=0;
	

	for( $i=0; $i<=$#$ATOM_aryref; $i++)
	{

		if($heavy_only eq "heavy" && $$ATOM_aryref[$i][4] eq "H")
		{
			next;
		}

		$mass_of=$Parameter::atomic_mass{$$ATOM_aryref[$i][4]};
		$x_sum+=$$ATOM_aryref[$i][9]*$mass_of;
		$y_sum+=$$ATOM_aryref[$i][10]*$mass_of;
		$z_sum+=$$ATOM_aryref[$i][11]*$mass_of;
		$mass_sum+=$mass_of;
	}
	$com_x=$x_sum/$mass_sum;
	$com_y=$y_sum/$mass_sum;
	$com_z=$z_sum/$mass_sum;
	return($com_x,$com_y,$com_z,$mass_sum);
}


return(1);
