############################################################################
#                                                                          #
#   Copyright (C) 2011 Po-Hsien Lee <p.lee@mx.uni-saarland.de>             #
#   All rights reserved.                                                   #
#                                                                          #
#   This file is part of PROPORES.                                         #
#                                                                          #
#   PROPORES is free software: you can redistribute it and/or modify       #
#   it under the terms of the GNU General Public License as published by   #
#   the Free Software Foundation, either version 3 of the License, or      #
#   (at your option) any later version.                                    #
#                                                                          #
#   PROPORES is distributed in the hope that it will be useful,            #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of         #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
#   GNU General Public License for more details.                           #
#                                                                          #
#   You should have received a copy of the GNU General Public License      #
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.  #
#                                                                          #
############################################################################
#!/usr/bin/perl

package Quat_rot;

sub new
{
        my $type = $_[0];
        my $self = {};
        bless($self, $type);
        return($self);
}

$pi=3.14159265359;

### rotate side chain of a residue
sub side_chain_rotate
{
	my $ref_di_def=$_[1]; my $side_chain_aryref=$_[2]; my $angle=$_[3];
	my $v1; my $v2; my $v3; my @rotM;
	my $length; my $start_index=0; my @ref_atoms=();


	$length=$#{$side_chain_aryref};	
	for(my $i=0; $i<=$length; $i++)
	{
		if($$side_chain_aryref[$i][3] eq $$ref_di_def[1])
		{
			@{$ref_atoms[0]}=@{$$side_chain_aryref[$i]}[9..11];
		}
		elsif($$side_chain_aryref[$i][3] eq $$ref_di_def[2])
		{
			@{$ref_atoms[1]}=@{$$side_chain_aryref[$i]}[9..11];
			$start_index=$i;
		}
	}
	

	$v1=$ref_atoms[0][0]-$ref_atoms[1][0];
	$v2=$ref_atoms[0][1]-$ref_atoms[1][1];
	$v3=$ref_atoms[0][2]-$ref_atoms[1][2];
	#print "$v1\t$v2\t$v3\t$angle\n";
	Quat_rot->quad_to_R($v1,$v2,$v3,$angle,\@rotM);
	Quat_rot->coor_rotate(\@rotM,$side_chain_aryref,$start_index);
}


### compute quaternion according the angle and then converse the quaternion to rotational matrix
sub quad_to_R
{
	my $v1=$_[1]; my $v2=$_[2]; my $v3=$_[3];
	my $angle=$_[4]; my $omega; my $norm;
	my $q1; my $q2; my $q3; my $q4;
	my $t2; my $t3; my $t4; my $t5; my $t6; my $t7; my $t8; my $t9; my $t10;
	my $Rotate_aryref=$_[5];


	$omega=$angle/180/2*$pi;
	$norm=Quat_rot->norm_vec($v1,$v2,$v3);
	#print "$norm\t$omega\n";

	$q1=cos($omega);
	$q2=$v1/$norm*sin($omega);
	$q3=$v2/$norm*sin($omega);
	$q4=$v3/$norm*sin($omega);

	$t2=$q1*$q2;
	$t3=$q1*$q3;
	$t4=$q1*$q4;
	$t5=-1*$q2**2;
	$t6=$q2*$q3;
	$t7=$q2*$q4;
	$t8=-1*$q3**2;
	$t9=$q3*$q4;
	$t10=-1*$q4**2;

	$$Rotate_aryref[0][0]=$t8+$t10; $$Rotate_aryref[0][1]=$t6-$t4; $$Rotate_aryref[0][2]=$t3+$t7;
	$$Rotate_aryref[1][0]=$t4+$t6; $$Rotate_aryref[1][1]=$t5+$t10; $$Rotate_aryref[1][2]=$t9-$t2;
	$$Rotate_aryref[2][0]=$t7-$t3; $$Rotate_aryref[2][1]=$t2+$t9; $$Rotate_aryref[2][2]=$t5+$t8;
}


### rotate coordinates according to the rotational matrix
sub coor_rotate
{
	my $Rotate_aryref=$_[1]; my $orig_coor_aryref=$_[2]; my $start_index=$_[3];
	my $x_new=0; my $y_new=0; my $z_new=0; my $length=0;
	my $tempx; my $tempy; my $tempz;

	$length=$#{$orig_coor_aryref};
	for(my $i=$start_index+1; $i<=$length; $i++)
	{
		$tempx=$$orig_coor_aryref[$i][9]-$$orig_coor_aryref[$start_index][9];
		$tempy=$$orig_coor_aryref[$i][10]-$$orig_coor_aryref[$start_index][10];
		$tempz=$$orig_coor_aryref[$i][11]-$$orig_coor_aryref[$start_index][11];
		$x_new=2*($$Rotate_aryref[0][0]*$tempx+$$Rotate_aryref[0][1]*$tempy+$$Rotate_aryref[0][2]*$tempz)+$tempx;
		$y_new=2*($$Rotate_aryref[1][0]*$tempx+$$Rotate_aryref[1][1]*$tempy+$$Rotate_aryref[1][2]*$tempz)+$tempy;
		$z_new=2*($$Rotate_aryref[2][0]*$tempx+$$Rotate_aryref[2][1]*$tempy+$$Rotate_aryref[2][2]*$tempz)+$tempz;

		$$orig_coor_aryref[$i][9]=$x_new+$$orig_coor_aryref[$start_index][9];
		$$orig_coor_aryref[$i][10]=$y_new+$$orig_coor_aryref[$start_index][10];
		$$orig_coor_aryref[$i][11]=$z_new+$$orig_coor_aryref[$start_index][11];
	}
}


### compute the norm of the vector
sub norm_vec
{
	my $length=$#_; my $sum=0;
	for(my $i=1; $i<=$length; $i++)
	{
		#print "$_[$i]\n";
		$sum+=$_[$i]**2;
	}
	$sum=sqrt($sum);
	return ($sum);
}

return(1);

