############################################################################
#                                                                          #
#   Copyright (C) 2011 Po-Hsien Lee <p.lee@mx.uni-saarland.de>             #
#   All rights reserved.                                                   #
#                                                                          #
#   This file is part of PROPORES.                                         #
#                                                                          #
#   PROPORES is free software: you can redistribute it and/or modify       #
#   it under the terms of the GNU General Public License as published by   #
#   the Free Software Foundation, either version 3 of the License, or      #
#   (at your option) any later version.                                    #
#                                                                          #
#   PROPORES is distributed in the hope that it will be useful,            #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of         #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
#   GNU General Public License for more details.                           #
#                                                                          #
#   You should have received a copy of the GNU General Public License      #
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.  #
#                                                                          #
############################################################################
#!/usr/bin/perl

package Vector;

use Math::Trig;
use POSIX;

$pi=3.14159265359;
$rad2deg=180/$pi;


sub new
{
        my $type = $_[0];
        my $self = {};
        bless($self, $type);
        return($self);
}


### compute the cross vector of two vectors in 3D
sub Cross_3D
{
	my $x1=$_[1]; my $x2=$_[2]; my $x3=$_[3];
	my $y1=$_[4]; my $y2=$_[5]; my $y3=$_[6];
	my $o1; my $o2; my $o3;

	$o1=$x2*$y3-$x3*$y2;
	$o2=$x3*$y1-$x1*$y3;
	$o3=$x1*$y2-$x2*$y1;

	return($o1,$o2,$o3);
}


### compute the dot value of two vectors
sub vec_dot
{
	my $aryref1=$_[1]; my $aryref2=$_[2];
	my $length=$#{$aryref1}; my $sum=0;

	for(my $i=0; $i<=$length; $i++)
	{
		$sum+=$$aryref1[$i]*$$aryref2[$i];
		#print "$$aryref1[$i] * $$aryref2[$i] +";
	}
	#print " $sum\n";
	return($sum);
}


### square of the norm of a vector
sub norm_vec_square
{
	my $length=$#_; my $sum=0;
	for(my $i=1; $i<=$length; $i++)
	{
		$sum+=$_[$i]**2;
	}
	return ($sum);
}


### norm of a vector
sub norm_vec
{
	my $sum=0; my @array=@_;
	shift(@array);
	$sum=sqrt(Vector->norm_vec_square(@array));
	return ($sum);
}


### compute the dihedral angle in degree
sub dihedral
{
	my $cor1_aryref=$_[1]; my $cor2_aryref=$_[2]; my $cor3_aryref=$_[3]; my $cor4_aryref=$_[4];
	my @pv=(); my @v1=(); my @v2=(); my @v3=(); my $norm_v1=0; my $norm_v2=0;	my $angle;


	my $x1=$$cor1_aryref[0]; my $y1=$$cor1_aryref[1]; my $z1=$$cor1_aryref[2];
	my $x2=$$cor2_aryref[0]; my $y2=$$cor2_aryref[1]; my $z2=$$cor2_aryref[2];
	my $x3=$$cor3_aryref[0]; my $y3=$$cor3_aryref[1]; my $z3=$$cor3_aryref[2];
	my $x4=$$cor4_aryref[0]; my $y4=$$cor4_aryref[1]; my $z4=$$cor4_aryref[2];


	$pv[0]=$x3-$x2;
	$pv[1]=$y3-$y2;
	$pv[2]=$z3-$z2;

	($v1[0],$v1[1],$v1[2])=Vector->Cross_3D(($x2-$x1),($y2-$y1),($z2-$z1),$pv[0],$pv[1],$pv[2]);
        ($v2[0],$v2[1],$v2[2])=Vector->Cross_3D(($x4-$x3),($y4-$y3),($z4-$z3),-$pv[0],-$pv[1],-$pv[2]);

        $norm_v1=Vector->norm_vec($v1[0],$v1[1],$v1[2]);
        $norm_v2=Vector->norm_vec($v2[0],$v2[1],$v2[2]);

        $angle=($v1[0]*$v2[0]+$v1[1]*$v2[1]+$v1[2]*$v2[2])/($norm_v1*$norm_v2);
        if($angle > 1.0)
        {
                $angle = 1.0;
        }
        elsif($angle < -1.0)
        {
                $angle = -1.0;
        }


        $angle=acos($angle)*$rad2deg;

        ($v3[0],$v3[1],$v3[2])=Vector->Cross_3D($v1[0],$v1[1],$v1[2],$v2[0],$v2[1],$v2[2]);

        if(Vector->vec_dot(\@v3,\@pv) < 0)
        {
                $angle*=-1;
        }
	
	return($angle);
}


### angle composed by two vectore in degree
sub angle
{
	my (@v1,@v2);
	$v1[0]=$_[1]; $v1[1]=$_[2]; $v1[2]=$_[3];
	$v2[0]=$_[4]; $v2[1]=$_[5]; $v2[2]=$_[6];
	my $norm_v1=Vector->norm_vec($v1[0],$v1[1],$v1[2]);
	my $norm_v2=Vector->norm_vec($v2[0],$v2[1],$v2[2]);


	my $angle=Vector->vec_dot(\@v1,\@v2)/($norm_v1*$norm_v2);
	if($angle > 1.0)
	{
		$angle = 1.0;
	}
	elsif($angle < -1.0)
	{
		$angle = -1.0;
	}

	$angle=acos($angle)*$rad2deg;
	
	return($angle);
}


### square of cosine value of two vectors
sub cos_sq
{
	my (@v1,@v2);
	$v1[0]=$_[1]; $v1[1]=$_[2]; $v1[2]=$_[3];
	$v2[0]=$_[4]; $v2[1]=$_[5]; $v2[2]=$_[6];
	my $norm_sq_v1=Vector->norm_vec_square($v1[0],$v1[1],$v1[2]);
	my $norm_sq_v2=Vector->norm_vec_square($v2[0],$v2[1],$v2[2]);


	my $product=(Vector->vec_dot(\@v1,\@v2))**2/($norm_sq_v1*$norm_sq_v2);

	return($product);
}


### copy multidimensional array
sub copy_MD_array
{
	shift(@_);
	my ($target_aryref,$original_aryref)=@_;
	my ($iter, $i);

	$iter=$#{$original_aryref};
	for($i=0; $i<=$iter; $i++)
	{
		if($$original_aryref[$i]=~m/^ARRAY\(\w+\)/)
		{
			Vector->copy_MD_array(\@{$$target_aryref[$i]},\@{$$original_aryref[$i]});
		}
		else
		{
			$$target_aryref[$i]=$$original_aryref[$i];
		}
	}
}


return(1);
