############################################################################
#                                                                          #
#   Copyright (C) 2011 Po-Hsien Lee <p.lee@mx.uni-saarland.de>             #
#   All rights reserved.                                                   #
#                                                                          #
#   This file is part of PROPORES.                                         #
#                                                                          #
#   PROPORES is free software: you can redistribute it and/or modify       #
#   it under the terms of the GNU General Public License as published by   #
#   the Free Software Foundation, either version 3 of the License, or      #
#   (at your option) any later version.                                    #
#                                                                          #
#   PROPORES is distributed in the hope that it will be useful,            #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of         #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          #
#   GNU General Public License for more details.                           #
#                                                                          #
#   You should have received a copy of the GNU General Public License      #
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.  #
#                                                                          #
############################################################################
#                                                                          #
#   shared_lining_residues.pl                                              #
#                                                                          #
#   Programmed by Po-Hsien Lee <p.lee@mx.uni-saarland.de>, 2011            #
#   Center for Bioinformatics, Saarland University, Germany                #
#                                                                          #
#   shared_lining_residues.pl is a program for figuring out the shared     #
#   lining residues between pores identified by Pore_ID.pl.                #
#                                                                          #
############################################################################

#!/usr/bin/perl

### check the input parameters
### usage of this program
if($#ARGV != 1)
{
	print "\nusage : perl shared_lining_residues.pl <prefix> <work_path>\n\n";
	print "\tThe <prefix> is the common prefix of the input (prefix)_#.list files\n";
	print "\tThe <work_path> is the absolute path of directory where (prefix)_#.list files exist\n\n";
	exit;
}
$output_prefix=$ARGV[0];
$work_path=$ARGV[1];

opendir DIR ,"$work_path" or die "$!";
@list=readdir(DIR);
closedir(DIR);


### read in all pore lining residues identified by Pore_ID.pl
@total_info=();
for($i=0;$i<=$#list;$i++)
{
	#print "$list[$i]\n";
	if($list[$i]=~m/$output_prefix\_(\d+)\.list/)
	{
		$index=$1; $count=0;
		open FH, "<$list[$i]" or die "$!";
		while(<FH>)
		{
			if(m/(.{1})\s+(\d+)\s+(\w{3})/)
			{
				$total_info[$index][$count][0]=$1;
				$total_info[$index][$count][1]=$2;
				$total_info[$index][$count][2]=$3;
				$count++;
			}
		}
		close FH;
	}
}


### compare and output shared pore-lining residues
for($k=0; $k<=$#total_info-1; $k++)
{
	%largest_set=();
	$length=$#{$total_info[$k]};
	for($i=0; $i<=$length; $i++)
	{
		$largest_set{$total_info[$k][$i][0]}{$total_info[$k][$i][1]}{$total_info[$k][$i][2]}=1;
	}


	for($i=$k+1; $i<=$#total_info; $i++)
	{
		$flag=0; $output_script="";
		$length2=$#{$total_info[$i]};
		for($j=0; $j<=$length2; $j++)
		{
			$chain_idx=$total_info[$i][$j][0];
			$residue_idx=$total_info[$i][$j][1];
			$residue_type=$total_info[$i][$j][2];
			if($largest_set{$chain_idx}{$residue_idx}{$residue_type}==1)
			{
				$flag=1;
				$output_script.="$chain_idx\t$residue_idx\t$residue_type\n";
			}
		}

		if($flag==1)
		{
			open FH, ">gating_residues_$output_prefix\_$k\_$i.info" or die "$!";
			print FH "$output_script";
			close FH;
		}
	}
}
